import copy

from marshmallow import Schema, pre_load, ValidationError
from .transform_fields import NestedValueField
from .utils import get_value_from_field, NotFoundValue
import typing


class ActionWhenValidateErrorField:
    REMOVE_FIELD = 'REMOVE_FIELD'


class BaseSchemaTransform(Schema):
    def __init__(
            self,
            *args,
            action_when_validate_error: ActionWhenValidateErrorField or None = None,
            **kwargs
    ):
        super(BaseSchemaTransform, self).__init__(*args, **kwargs)
        self.action_when_validate_error_field = action_when_validate_error

    def _get_value_for_nested_value_field(self, data):
        for attr_name, field_obj in self.load_fields.items():
            if isinstance(field_obj, NestedValueField):
                field_name = (
                    field_obj.nested_key if field_obj.nested_key is not None else attr_name
                )
                try:
                    data[attr_name] = get_value_from_field(data, field_name)
                except NotFoundValue as e:
                    pass

    def _return_type_class_for_nested_value_field(self):
        load_fields_copy = copy.deepcopy(self.load_fields)
        for attr_name, field_obj in self.load_fields.items():
            if isinstance(field_obj, NestedValueField) and field_obj.type_class is not None:
                attrs_of_field_obj = copy.deepcopy(field_obj.__dict__)
                load_fields_copy[attr_name] = field_obj.type_class(**attrs_of_field_obj)
        self.load_fields = load_fields_copy

    def pre_load_action(self, data, many, **kwargs):
        return data

    @pre_load(pass_many=True)
    def _pre_load_action(self, data: dict, many, **kwargs):
        data = self.pre_load_action(data, many, **kwargs)
        data_list = data
        if not many:
            data_list = [data]

        data_list_copy = []
        for i in range(0, len(data_list)):
            data_item = data_list[i]
            self._get_value_for_nested_value_field(data_item)
            data_list_copy.append(data_item)
        self._return_type_class_for_nested_value_field()

        if not many:
            return data_list_copy[0]
        return data_list_copy

    def remove_fields_error(self, data, fields_error):
        for field in fields_error:
            del data[field]

    def load(
        self,
        data: (
            typing.Mapping[str, typing.Any]
            or typing.Iterable[typing.Mapping[str, typing.Any]]
        ),
        *,
        many: bool or None = None,
        partial: bool or None = None,
        unknown: str or None = None,
    ):
        try:
            return super().load(data, many=many, partial=partial, unknown=unknown)
        except ValidationError as e:
            error_messages: typing.Mapping[int, dict] | typing.Mapping[str, list] = e.messages
            data_list = data

            if not many:
                error_messages = {0: error_messages}
                data_list = [data]

            if self.action_when_validate_error_field is not None:
                if self.action_when_validate_error_field == ActionWhenValidateErrorField.REMOVE_FIELD:
                    for index, fields_error in error_messages.items():
                        self.remove_fields_error(data_list[index], list(fields_error.keys()))
                    return super().load(data, many=many, partial=partial, unknown=unknown)
            raise e

    @classmethod
    def transform(
        cls,
        data: (
            typing.Mapping[str, typing.Any]
            or typing.Iterable[typing.Mapping[str, typing.Any]]
        ),
        *,
        many: bool or None = False,
        partial: bool or None = False,
        unknown: str or None = None,
        action_when_validate_error: ActionWhenValidateErrorField or None = None
    ):
        data_copy = copy.deepcopy(data)
        instance = cls(many=many, partial=partial, unknown=unknown, action_when_validate_error=action_when_validate_error)
        return instance.load(data_copy, many=many, partial=partial, unknown=unknown)

    @classmethod
    def get_require_fields(cls):
        self = cls()
        require_fields = {}
        for attr_name, field_obj in self.load_fields.items():
            if field_obj.required:
                require_fields[attr_name] = field_obj
        return require_fields

    def dump(self, obj: typing.Any, *, many: bool or None = None):
        raise Exception('Method not use.')

    def dumps(self, obj: typing.Any, *args, many: bool or None = None, **kwargs):
        raise Exception('Method not use')



import copy
import datetime


def convert_timestamp_to_date_utc(timestamp):
    try:
        if timestamp is not None:
            return datetime.datetime.utcfromtimestamp(timestamp).replace(
                tzinfo=datetime.timezone.utc
            )
        else:
            return None
    except Exception as e:
        return None


class NoValue:
    pass


class NotFoundValue(Exception):
    pass


class DataIsNotDict(Exception):
    pass


def is_atom_value(value):
    return value is None or value == {} or not isinstance(value, dict)


def get_value_from_field(data: dict, from_field: str, default_value=NoValue()):
    data_copy = copy.deepcopy(data)
    field_road = from_field.split('.')

    value = NoValue()
    for field in field_road:
        try:
            value = data_copy[field]
        except Exception:
            if not isinstance(default_value, NoValue):
                value = default_value
                break
            raise NotFoundValue('Not found value')

        data_copy = value

    if isinstance(value, NoValue):
        raise NotFoundValue('Not found value')
    return value


def remove_none_field(data) -> (dict, dict):
    data_copy = copy.deepcopy(data)
    none_field_map = copy.deepcopy(data)

    def recurse_dict(data_in, data_copy_in, none_field_map_in, none_field_map_in_before=None, key_before=None):
        for key, value in data_in.items():
            if value is None:
                del data_copy_in[key]
            elif (value is not None and not isinstance(value, dict)) or value == {}:
                del none_field_map_in[key]
            elif isinstance(value, dict):
                recurse_dict(value, data_copy_in[key], none_field_map_in[key], none_field_map_in, key)

        if none_field_map_in == {} and none_field_map_in_before is not None:
            del none_field_map_in_before[key_before]

    recurse_dict(data, data_copy, none_field_map)
    return data_copy, none_field_map


def merge_none_field_map_into_data_without_none_field(none_field_map: dict, data_without_none_field: dict):
    for key, value in none_field_map.items():
        if not isinstance(value, dict) and value is None:
            data_without_none_field[key] = value
        elif isinstance(value, dict):
            merge_none_field_map_into_data_without_none_field(value, data_without_none_field[key])



from marshmallow import fields, ValidationError, utils
import typing
from .utils import convert_timestamp_to_date_utc
from .auto_replace_validate_error_field import AutoReplaceErrorField
import datetime


__all__ = [
    "NestedValueField",
    "DatetimeFromTimeStamp"
]


class NestedValueField(fields.Field, metaclass=AutoReplaceErrorField):
    """Field that help to get value from nested field when deserializer
    :param nested_key: show where to get value in nested field"""

    def __init__(self, *args, **kwargs):
        if 'data_key' in kwargs and 'nested_key' in kwargs:
            raise Exception('You must define just one of data_key or nested_key.')
        super(NestedValueField, self).__init__(*args, **kwargs)
        self.type_class = kwargs.get('type_class', None)
        self.nested_key = kwargs.get('nested_key', None)


class Str(fields.Str, metaclass=AutoReplaceErrorField):
    pass


class Int(fields.Int, metaclass=AutoReplaceErrorField):
    pass


class Float(fields.Float, metaclass=AutoReplaceErrorField):
    pass


class Date(fields.Date, metaclass=AutoReplaceErrorField):
    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, datetime.date):
            return value
        else:
            return super(Date, self)._deserialize(value, attr, data, **kwargs)


class DateTime(fields.DateTime, metaclass=AutoReplaceErrorField):
    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, datetime.datetime):
            return value
        else:
            return super(DateTime, self)._deserialize(value, attr, data, **kwargs)


class Nested(fields.Nested, metaclass=AutoReplaceErrorField):
    pass


class DatetimeFromTimeStamp(fields.Field):
    """Field that deserializer from timestamp like: 1660622040.0 to datetime object"""

    def _deserialize(
            self,
            value: typing.Any,
            attr: str or None,
            data: typing.Mapping[str, typing.Any] or None,
            **kwargs,
    ):
        if value is None:
            return None
        return convert_timestamp_to_date_utc(value)



import typing
from marshmallow import ValidationError


def super_call(supers):
    if len(supers) >= 1:
        return supers[0]
    else:
        return object


def add_hook_into_init(cls, supers):
    original_init = cls.__init__

    def __init__(self, *args, **kwargs):
        original_init(self, *args, **kwargs)
        self.value_when_validate_error_hook = {
            'set': False
        }
        if 'value_when_validate_error' in kwargs:
            self.value_when_validate_error_hook = {
                'set': True,
                'value': kwargs.get('value_when_validate_error')
            }
    cls.__init__ = __init__


def add_hook_into_deserialize(cls, supers):
    def deserialize(
        self,
        value: typing.Any,
        attr: str or None = None,
        data: typing.Mapping[str, typing.Any] or None = None,
        **kwargs,
    ):
        try:
            deserialized_value = super_call(supers).deserialize(self, value, attr, data, **kwargs)
            return deserialized_value
        except ValidationError as e:
            if self.value_when_validate_error_hook['set']:
                return self.value_when_validate_error_hook['value']
            else:
                raise e
    cls.deserialize = deserialize


class AutoReplaceErrorField(type):
    def __init__(cls, class_name, supers, classdict):
        add_hook_into_init(cls, supers)
        add_hook_into_deserialize(cls, supers)



from .base_schema import BaseSchemaTransform, ActionWhenValidateErrorField
from . import transform_fields
from marshmallow import *
from marshmallow import validate

__all__ = [
    "BaseSchemaTransform",
    "transform_fields",
    'ActionWhenValidateErrorField',
    "EXCLUDE",
    "INCLUDE",
    "RAISE",
    "Schema",
    "SchemaOpts",
    "validates",
    "validates_schema",
    "pre_dump",
    "post_dump",
    "pre_load",
    "post_load",
    "pprint",
    "ValidationError",
    "missing",
    "validate"
]
